/*
 * This file is part of dial.
 *
 * dial is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * dial is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dial; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Copyright 1994 Dominic Giampaolo
 *   Copyright 1996 Torsten Martinsen
 *
 * $Id: connect.c,v 1.17 1996/03/27 12:08:05 torsten Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>

#include "dial.h"
#include "readrc.h"

static int * cstatep;
static pid_t child_pid;
static int aborting = 0;

static void checklink(void * timearg);
static void do_wait(void * func);
static void redial(void * junk);
static void do_redial(void * junk);

char * number = NULL;

/*
 * Start the PPP daemon and set up a routine to wait until the link is up.
 */
void
doconnect(int * connectp)
{
  pid_t pid;
  char command[80];
  
  cstatep = connectp;

  *cstatep = CSTATE_TRYING;

#ifdef TEST
  if (!testing) {
#endif
  pid = fork();
  if (!pid)
    {
      sprintf(command, "%s %s", CONNECTSCRIPT, number);
      execl(PPPD, PPPD, "connect", command, NULL);
      printf("execl() failed\n");
    }
  else if (pid == -1)
    printf("fork() failed\n");
  else
    child_pid = pid;

  SetLabel(button, "Abort");
  SetWidgetState(button, TRUE);
  aborting = 0;

  /*
   * Wait a realistic amount of time before calling checklink()
   * the first time.
   */
  AddTimeOut(5000, checklink, (void *) time(NULL));
#ifdef TEST
  } else {
    *cstatep = CSTATE_CONNECTED;
    add_dial_cost();
    startclock();
    set_display_state(DS_CONNECTED);
    SetLabel(button, "Disconnect");
    SetWidgetState(button, TRUE);
  }
#endif
}

#ifdef SHOW_CONNECT_STATUS
/*
 * The hairy version. Periodically check the syslog file to gather
 * messages from pppd.
 */
static void
checklink(void * timearg)
{
  static time_t start, now;
  static size_t fpos;
  FILE * fp;
  char s[80], *p;
  static int dialed = 0;
  int done = 0;
  
  if (aborting)
    return;

  fp = fopen(SYSLOGFILE, "r");
  
  if (timearg)
    {
      start = (time_t) timearg;
      /* Get current size of log file */
      fseek(fp, 0, SEEK_END);
      fpos = ftell(fp);
    }

  /* Seek to the old end and see if any stuff has been added */
  fseek(fp, fpos, SEEK_SET);
  if (fgets(s, sizeof(s)-1, fp) != NULL)
    {
      /* We got a line; see if it is from chat */
      if ((p = strstr(s, "chat[")) != NULL)
	{
	  /* Check for some possible chat messages */
	  if (strstr(p, "ATD"))			/* Dial command */
	    dialed = 1;
	  else if (strstr(p, "Failed"))		/* Some error occurred */
	    {
	      if (strstr(p, "BUSY"))
		{
		  set_display_state(DS_BUSY);
		  if (global.redialinterval > 0)
		    {
		      *cstatep = CSTATE_BUSY;
		      AddTimeOut(3000, redial, NULL);
		    }
		  else
		    *cstatep = CSTATE_NC;
		}
	      else
		{
		  set_display_state(DS_CONNECT_FAILED);
		  *cstatep = CSTATE_NC;
		}
	      if (dialed)
		{
		  add_dial_cost();
		  dialed = 0;
		}
	      hangup(hangup_done);
	      done = 1;
	    }
	  else if (strstr(p, "NO DIALTONE"))
	    dialed = 0;
	}
      /* See if the message is from pppd */
      else if ((p = strstr(s, "pppd[")) != NULL)
	{
	  /* Check for some possible pppd messages */
	  if (strstr(p, "connection established"))
	    {
	      set_display_state(DS_SER_CONN_EST);
	    }
	  else if (strstr(p, "IP address"))
	    {
	      *cstatep = CSTATE_CONNECTED;
	      add_dial_cost();
	      startclock();
	      set_display_state(DS_CONNECTED);
	      SetLabel(button, "Disconnect");
	      done = 1;
	    }
	}
      fpos = ftell(fp);
    }
  fclose(fp);

  /* Check for timeout */
  now = time(NULL);
  if (now-start > global.connecttimeout)
    {
      set_display_state(DS_CONNECT_TIMED_OUT);
      *cstatep = CSTATE_NC;
      hangup(hangup_done);
    }
  else if (!done)
    /* Do it again, please */
    AddTimeOut(500, checklink, NULL);
}
#else
/*
 * The no-frills version. Periodically check for existence
 * of the 'linkinfo' file.
 */
static void
checklink(void * timearg)
{
  static time_t start, now;
  struct stat buf;
  static size_t fpos;
  FILE * fp;
  char s[80];

  
  if (aborting)
    return;

  if (timearg)
    start = (time_t) timearg;

  now = time(NULL);
  
  /* Check if the ip-up script says the link is up */
  if (stat(PPPDFILE, &buf) == 0)
    {
      *cstatep = CSTATE_CONNECTED;
      startclock();
      set_display_state(DS_CONNECTED);
      SetLabel(button, "Disconnect");
    }
  else if (now-start > global.connecttimeout)
    {
      set_display_state(DS_CONNECT_TIMED_OUT);
      *cstatep = CSTATE_NC;
      hangup(hangup_done);
    }
  else
    AddTimeOut(1000, checklink, NULL);
}
#endif

/*
 * Hang up the modem by sending a SIGINT to the PPP daemon.
 */
void
hangup(Voidfunc func)
{
  FILE * fp;
  char buf[15];
  pid_t pid;

  log_costs();
  if ((*cstatep == CSTATE_TRYING) || (*cstatep == CSTATE_BUSY))
    aborting = 1;
  
  /* Get the pid from the /var/run/ppp0.pid file */
  if ((fp = fopen(PPPDPIDFILE, "r")) == NULL)
    {
      child_pid = 0;
    }
  else
    {
      fgets(buf, sizeof(buf)-1, fp);
      fclose(fp);
      pid = atoi(buf);
      if (!pid)
	printf("error getting pid\n");
      kill(pid, SIGINT);
    }
  AddTimeOut(1000, do_wait, func);
}

static void
do_wait(void * func)
{
  if (child_pid)
    wait(NULL);

  ((Voidfunc) func)();
}

/*
 * Let the status area show that redialing will happen.
 * Then set up a timeout to actually do it.
 */
static void
redial(void * junk)
{
  set_display_state(DS_WAITING);
  AddTimeOut(global.redialinterval*1000, do_redial, NULL);
}

static void
do_redial(void * junk)
{
  doconnect(cstatep);
}
