/*
 * This file is part of dial.
 *
 * dial is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * dial is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dial; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Copyright 1994 Dominic Giampaolo
 *   Copyright 1996 Torsten Martinsen
 *
 * $Id: cost.c,v 1.15 1996/03/27 18:30:39 torsten Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/param.h>
#include <pwd.h>
#include <sys/types.h>

#include "dial.h"
#include "readrc.h"

static time_t starttime, stoptime;
static int running;
static float cost;
static int inited = 0;
static int callstarted;

static char clockstring[10], coststring[10], ratestring[10];

static float getrate(time_t now);
static int isholiday(struct tm * lt);

typedef struct
{
  int day;
  int month;
} hday;

/*
 * Called every second.
 * If we are not connected, just update rate.
 * Otherwise, also update time and cost.
 */
void
clocktick(void * junk)
{
  time_t now = time(NULL);
  time_t diff = now-starttime, temp;
  int h, m, s, ratechanged = 0, costchanged = 0, timechanged;
  static float rate = 0.0;
  static time_t lastupdate;
  float t;
  
  h = diff/3600;
  temp = diff - h*3600;
  m = diff/60;
  s = diff - m*60;

  timechanged = running;
  if (!inited)
    {
      strcpy(clockstring, "0:00:00");
      rate = getrate(now);
      ratechanged = 1;
      costchanged = 1;
      lastupdate = now;
      inited = 1;
      timechanged = 1;

    }
  if ((t = now-lastupdate) >= COSTINTERVAL)
    {
      float oldrate = rate;
      
      rate = getrate(now);
      ratechanged = (rate != oldrate);
      if (running)
	{
	  cost += rate*t/60.0;
	  costchanged = 1;
	}
      lastupdate = now;
    }

  if (running)
    sprintf(clockstring, "%2d:%02d:%02d", h, m, s);
  if (costchanged)
    sprintf(coststring, "%.2f", cost);
  if (ratechanged)
    sprintf(ratestring, "%.2f", rate);
  set_display(clockstring, coststring, ratestring);
  update_display(timechanged, costchanged, ratechanged);
  
  AddTimeOut(TICKINTERVAL, clocktick, NULL);
}

void
startclock(void)
{
  starttime = time(NULL);
  running = 1;
  inited = 0;
}

void
stopclock(void)
{
  stoptime = time(NULL);
  running = 0;
}

void
add_dial_cost(void)
{
  cost = global.dialcost;
  callstarted = 1;
  starttime = time(NULL);
  stoptime = 0;
}

static float
getrate(time_t now)
{
  struct tm * local;
  float hi, lo, rate;
  Whennight * wp;
  int s, e, n;
  
  local = localtime(&now);
  
  lo = curnuminfo->nightrate;
  hi = curnuminfo->dayrate;

  rate = hi;

  if (curnuminfo->cheapdays[local->tm_wday])
    rate = lo;
  else if (isholiday(local))
    rate = lo;
  else
    {
      n = local->tm_hour*60+local->tm_min;
      for (wp = curnuminfo->whennight; wp != NULL; wp = wp->next)
	{
	  s = wp->s_hour*60+wp->s_min;
	  e = wp->e_hour*60+wp->e_min;
	  if ((n >= s) && (n < e))
	    return lo;
	}
    }

  return rate;
}

static hday holidays[] = {
  {  1,  0 },	/* New Year's Day */
  { 24, 11 },	/* Cristmas Eve */
  { 25, 11 },	/* Cristmas Day */
  { 26, 11 },
  { -1, -1 }
};

/*
 * Return 1 if the specified day is a holiday.
 * For now, only fixed holidays are implemented.
 * If you know a funky algorithm to take Easter etc. into account,
 * let me know.
 */
static int
isholiday(struct tm * lt)
{
  hday * hd = holidays;
  
  if (lt->tm_wday == 0) /* Sunday */
    return 1;
  while (hd->month >= 0)
    {
      if ((hd->month == lt->tm_mon) && (hd->day == lt->tm_wday))
	return 1;
      ++hd;
    }
  return 0;
}
  
/*
 * Append a line describing the just-ended session to the log file.
 */
void
log_costs(void)
{
  FILE * fp;
  time_t diff;
  int h, m, s;
  char start[30], stop[30];


  if (!callstarted)
    return;
  
  if ((fp = fopen(global.logfile, "a")) == NULL)
    {
      char buf[MAXPATHLEN];

      sprintf(buf, "can't open %s", global.logfile);
      perror(buf);
    }
  else
    {
      /* Get start time string without newline */
      strcpy(start, asctime(localtime(&starttime)));
      start[strlen(start)-1] = 0;
      if (stoptime)
	{
	  /* Elapsed time in h/m/s */
	  diff = stoptime-starttime;
	  h = diff/3600;
	  diff -= h*3600;
	  m = diff/60;
	  s = diff - m*60;

	  /* Get stop time string without newline */
	  strcpy(stop, asctime(localtime(&stoptime)));
	  stop[strlen(stop)-1] = 0;
	  
#ifdef TEST
	  if (testing)
	    fprintf(fp, "TEST ");
#endif
	  fprintf(fp, "%s-%s: Time %d:%02d:%02d  Cost %.2f\n", 
		  start, stop, h, m, s, cost);
	}
      else if (global.dialcost != 0.0)
	{
	  /* Call did not get through, log cost of dialling */
#ifdef TEST
	  if (testing)
	    fprintf(fp, "TEST ");
#endif
	  fprintf(fp,
		  "%s: Attempted call                         Cost %.3f\n", 
		  start, cost);
	}
    }
	  
  fclose(fp);
  callstarted = 0;
}
