/*
 * This file is part of dial.
 *
 * dial is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * dial is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dial; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 1996 Torsten Martinsen
 *
 * $Id: readrc.c,v 1.5 1996/03/27 18:33:59 torsten Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>

#include "dial.h"
#include "readrc.h"

GlobalOptions global = {
   60,  /* redialinterval */  
   60,  /* connecttimeout */  
    0,  /* redial */   
  0.0,  /* dialcost */
  NULL, /* logfile */      
  NULL  /* defaultnumber */
};

static void parseint(char *s, int *arg);
static void parseflt(char *s, float *arg);
static void parsestr(char * s, char ** arg);
static char *expand_tilde(char *path);
static NumberInfo * currentNumberInfo(void);
static void parseday(char *s);
static void parsenum(char *s);
static void parserdi(char *s);
static void parsecto(char * s);
static void parsedlc(char * s);
static void parsedrt(char * s);
static void parsenrt(char * s);
static void parsetime(char *s);
static void parselog(char *s);
static void parsedef(char *s);
static void rcerror(char * reason);

static int lineno;
static NumberInfo * numberinfocur, * numberinfohead = NULL;
static char * rcfile;

/* These must be in the same order that struct tm uses */
static char * weekdays[] = {
  "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};


static Keywords keywords[] = {
  /* Global parameters */
  { "redialinterval",	parserdi },
  { "connecttimeout",	parsecto },
  { "dialcost",		parsedlc },
  { "logfile",		parselog },
  { "defaultnumber",	parsedef },
  /* Per-number parameters */
  { "number",		parsenum },
  { "nightrate",	parsenrt },
  { "dayrate",		parsedrt },
  { "nighttime",	parsetime },
  { "cheapday",		parseday },
  { NULL }
};

void
parserc(char * file)
{
  char * fname, buf[200], * p, * cur, * w;
  FILE * fp;
  Keywords * kp;

  rcfile = file;
  fname = expand_tilde(rcfile);
  if ((fp = fopen(fname, "r")) == NULL)
    return;
  
  /* Read each line, skip empty or comment lines, and parse the line. */
  lineno = 0;
  while (fgets(buf, sizeof(buf)-1, fp))
    {
      ++lineno;
      p = buf;
      while (*p && isspace(*p))
	++p;
      if ((*p == '#') || (*p == '\n') || (*p == '\0'))
	continue;
      for (kp = keywords; kp->name != NULL; ++kp)
	{
	  cur = kp->name;
	  w = p;
	  while (*cur && *w && (*cur == *w))
	    {
	      ++cur;
	      ++w;
	    }
	  if ((*cur == 0) && isspace(*w))
	    {
	      while (*w && isspace(*w))
		++w;
	      kp->func(w);
	      break;
	    }
	}
      if (kp->name == NULL)
	rcerror("unknown keyword");
    }
  fclose(fp);

  /* Hack to get leading tilde expanded */
  if (global.logfile == NULL)
    parselog(DEFAULTLOGFILE);
  
#if 0
  printf("Parsed:\n"
	 "Redial interval      %d\n"
	 "Connect timeout      %d\n"
	 "Dial cost            %f\n"
	 "Logfile              %s\n"
	 "Default phone number %s\n",
	 global.redialinterval, global.connecttimeout,
	 global.dialcost, global.logfile, global.defaultnumber);
  {
    NumberInfo * nip;
    Whennight * p;
    int i, cd;

    for (nip = numberinfohead; nip != NULL; nip=nip->next)
      {
	printf("------------------\n"
	       "Phone number: %s\n"
	       "Day rate   %.2f\n"
	       "Night rate %.2f\n",
	       nip->number, nip->dayrate, nip->nightrate);
	for (p = nip->whennight; p != NULL; p = p->next)
	  printf("Night rate applies from %02d:%02d to %02d:%02d\n",
		 p->s_hour, p->s_min, p->e_hour, p->e_min);
	cd = 0;
	for (i = 0; i < 7; ++i)
	  if (nip->cheapdays[i])
	    cd = 1;
	if (cd || nip->cheapholidays)
	  {
	    printf("Night rate also applies on: ");
	    for (i = 0; i < 7; ++i)
	      if (nip->cheapdays[i])
		printf("%s ", weekdays[i]);
	    if (nip->cheapholidays)
	      printf("%sholidays", cd ? "and " : "");
	    printf("\n");
	  }
      }
  }
#endif
}

/*
 * Generic parse routines
 */
static void
parseint(char * s, int * arg)
{
  if (sscanf(s, "%d", arg) != 1)
    rcerror("invalid integer");
}

static void
parseflt(char * s, float * arg)
{
  if (sscanf(s, "%f", arg) != 1)
    rcerror("invalid floating point number");
}

/*
 * Parse a string, expanding any leading tilde.
 * The string is allocated by strdup(), and must therefore be free()d after use.
 */
static void
parsestr(char * s, char ** arg)
{
  char temp[200];

  if (sscanf(s, "%199s", temp) == 1)
    *arg = strdup(temp);
  else
    *arg = strdup("");
}

#if 0
static void
parseyn(char * s, int * arg)
{
  *arg = (*s == 'y') || (*s == 'Y');
}
#endif

/*
 * Keyword-specific parse routines.
 */

static void
parserdi(char * s)
{
  parseint(s, &global.redialinterval);
}

static void
parsecto(char * s)
{
  parseint(s, &global.connecttimeout);
}

static void
parsedlc(char * s)
{
  parseflt(s, &global.dialcost);
}

static void
parselog(char * s)
{
  char * temp;
  
  parsestr(s, &temp);
  global.logfile = strdup(expand_tilde(temp));
  free(temp);
}

static void
parsedef(char * s)
{
  parsestr(s, &global.defaultnumber);
}

static void
parsedrt(char * s)
{
  parseflt(s, &(currentNumberInfo()->dayrate));
}

static void
parsenrt(char * s)
{
  parseflt(s, &(currentNumberInfo()->nightrate));
}

/*
 * Read a time specification of the form hh:mm-hh:mm.
 */
static void
parsetime(char * s)
{
  int n, sh, sm, eh, em;
  Whennight * new, * cur, * prev;
  NumberInfo * nip;
  
  n = sscanf(s, "%d:%d-%d:%d", &sh, &sm, &eh, &em);
  if ((n != 4) ||
      (sh < 0) || (sh > 23) || (sm < 0) || (sm > 59) ||
      (eh < 0) || (eh > 23) || (em < 0) || (em > 59) ||
      (sh*60+sm > eh*60+em))
    {
      rcerror("invalid time specification");
      return;
    }
  nip = currentNumberInfo();
  new = malloc(sizeof(Whennight));
  new->next = NULL;
  new->s_hour = sh;
  new->s_min = sm;
  new->e_hour = eh;
  new->e_min = em;
  if (nip->whennight == NULL)
    nip->whennight = new;
  else
    {
      prev = cur = nip->whennight;
      while (cur != NULL)
	{
	  prev = cur;
	  cur = cur->next;
	}
      prev->next = new;
    }
}

static void
parseday(char * s)
{
  int i;
  NumberInfo * nip;
  
  nip = currentNumberInfo();
  if (!strncasecmp(s, "hol", 3))
    {
      nip->cheapholidays = 1;
      return;
    }
  for (i = 0; i < 7; ++i)
    if (!strncasecmp(s, weekdays[i], 3))
      {
	nip->cheapdays[i] = 1;
	return;
      }
  rcerror("unrecognized day");
}


/*
 * Parse a phone number and initialize a NumberInfo structure
 * in which to store the following info.
 * Insert the NumberInfo struct in the linked list so that longest
 * numbers are first.
 */
static void
parsenum(char * s)
{
  char * num;
  NumberInfo * new, * nip, * prev;
  int i;
  
  parsestr(s, (void *) &num);
  new = malloc(sizeof(NumberInfo));
  new->number = num;
  new->dayrate = 0.0;
  new->nightrate = 0.0;
  new->whennight = NULL;
  for (i = 0; i < 7; ++i)
    new->cheapdays[i] = 0;
  new->cheapholidays = 0;
  if (numberinfohead == NULL)
    {
      new->next = NULL;
      numberinfohead = new;
    }
  else
    {
      i = strlen(num);
      nip = prev = numberinfohead;
      if (strlen(nip->number) <= i)
	{
	  new->next = numberinfohead;	/* Insert at list head */
	  numberinfohead = new;
	}
      else				/* Insert in middle of list */
	{
	  while (nip && (strlen(nip->number) > i))
	    {
	      prev = nip;
	      nip = nip->next;
	    }
	  if (nip == NULL)
	    nip = prev;
	  new->next = nip->next;
	  nip->next = new;
	}
    }
  numberinfocur = new;
}

static NumberInfo *
currentNumberInfo(void)
{
  int i;
  
  if (numberinfohead == NULL)
    {
      numberinfohead = malloc(sizeof(NumberInfo));
      numberinfohead->number = "";
      numberinfohead->dayrate = 0.0;
      numberinfohead->nightrate = 0.0;
      numberinfohead->whennight = NULL;
      for (i = 0; i < 7; ++i)
	numberinfohead->cheapdays[i] = 0;
      numberinfohead->cheapholidays = 0;
      numberinfohead->next = NULL;
      numberinfocur = numberinfohead;
    }
  return numberinfocur;
}  

/*
 * Expand leading tilde in path name.
 */
static char *
expand_tilde(char *path)
{
  static char out[MAXPATHLEN];
  char name[80];
  char *pp, *cp;
  struct passwd *pw;

  if (*path != '~')
    return path;
  ++path;
  pp = path;
  cp = name;
  while (*pp && (*pp != '/'))
    *cp++ = *pp++;
  *cp = 0;
  if (name[0] == '\0')
    pw = getpwuid(getuid());
  else
    pw = getpwnam(name);
  if (pw == NULL)
    return path;
  strcpy(out, pw->pw_dir);
  strcat(out, pp);

  return out;
}

static void
rcerror(char * reason)
{
  fprintf(stderr, "error parsing '%s': %s in line %d\n",
	  rcfile, reason, lineno);
}

static NumberInfo fallback = {
  "", 0.0, 0.0, NULL, { 0, 0, 0, 0, 0, 0, 0 }, 0, NULL
};

NumberInfo *
getNumberInfo(char * number)
{
  NumberInfo * nip = numberinfohead;
  char * p, * s;
  
  while (nip != NULL)
    {
      p = nip->number;
      s = number;
      while (*s && *p && (*s == *p))
	{
	  ++s;
	  ++p;
	}
      if (*p == 0)
	return nip;
      nip = nip->next;
    }
  return &fallback;
}
